(*

                   Inertia/TMT Realtime 3D Rendering Engine
     Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED

*)

Unit Inertia;
{$I DEFINE.INC}

Interface

Uses
  Types;

{Misc functions }
Procedure InitInertia;
Procedure ShutDownInertia;
Procedure SetAspectRatio(ARatio : Single);

{Object functions }
Function  LoadGVO(Var Obj : VectorObject; Filename : String; FOfs : Longint) : Longint;
Function  AllocObject(Var Obj : VectorObject; NumVert, NumPoly : Longint) : Longint;
Function  CopyVectorObject(Var Source, Dest : VectorObject) : Longint;
Procedure InitVectorObject(Var Obj : VectorObject);
Procedure FreeVectorObject(Var Obj : VectorObject);
Procedure SetAbsoluteRotation(Var Obj : VectorObject; rx, ry, rz : Single);
Procedure SetDeltaRotation(Var Obj : VectorObject; dx, dy, dz : Single);
Procedure SetAbsoluteLocation(Var Obj : VectorObject; tx, ty, tz : Single);
Procedure SetDeltaLocation(Var Obj : VectorObject; dx, dy, dz : Single);
Procedure SetBaseColor(Var Obj : VectorObject; C : Byte);
Procedure SetPolygonColor(Var Obj : VectorObject; Poly : Longint; C : Byte);
Procedure SetObjectTexture(Var Obj : VectorObject; C : Longint);
Procedure SetPolygonTexture(Var Obj : VectorObject; Poly, C : Longint);
Procedure SetObjectPhongMap(Var Obj : VectorObject; C : Longint);
Procedure SetPolygonPhongMap(Var Obj : VectorObject; Poly, C : Longint);
Procedure SetRenderMask(Var Obj : VectorObject; Mask : Longint);
Procedure SetCull(Var Obj : VectorObject; Cull : Longint);

{View Functions }
Procedure SetFocalLength(Var View : ViewObject; Fl : Single);
Procedure SetFieldOfView(Var View : ViewObject; FOV : Single);
Procedure InitView(Var View : ViewObject);
Procedure FreeView(Var View : ViewObject);
Procedure GetObjectAsCamera(Var View : ViewObject; Obj : VectorObject);
Procedure SetViewFustrum(Var View : ViewObject; MinZ, MaxZ : Single);
Procedure SetViewPort(Var View : ViewObject; X1, Y1, X2, Y2 : Single; Update : Boolean);
Procedure SetViewLocation(Var View : ViewObject; X, Y, Z : Single);
Procedure SetFocusPoint(Var View : ViewObject; TargetX, TargetY, TargetZ, Roll : Single);

{Light Functions }
Procedure PushLightSource(X, Y, Z : Single);
Procedure PopLightSource;
Procedure RotateLightSource(Index : Longint; Xa, Ya, Za : Single);

{Render functions }
Procedure Render(VPage : Pointer; X, Y : Longint);
Procedure AddToRenderList(View : ViewObject; Var Obj : VectorObject);

{Table Functions }
Function  LoadIntensityTable(Filename : String; FOfs : Longint) : Longint;
Procedure FreeIntensityTable;
Function  LoadTransparencyTable(Filename : String; FOfs : Longint) : Longint;
Procedure FreeTransparencyTable;
Function  LoadPalette(FileName : String; FOfs : Longint) : Longint;
Function  LoadPCXTexture(Index : Longint; Filename : String; FOfs : Longint) : Longint;
Procedure FreeTexture(Index : Longint);
Procedure GetTexturePalette(Index : Longint);
Function  MakePhongMap(Index : Longint) : Longint;

Implementation

Uses
  Misc,     { Miscelanous functions     }
  Shade,    { Shader functions          }
  Polygon,  { Polygon fillers           }
  Memory,   { Memory management         }
  Matrix,   { Matrix functions          }
  CFile,    { C-like file handler       }
  Clip,     { 3d clipping functions     }
  Vector,   { Vector functions          }
  Xform,    { 3d and 3d->2d transforms  }
  Cull,     { Object culling            }
  Pcx,      { PCX loading               }
  Pool;     { Vertex and Polygon stores }

{===========================================================================}
{ InitInertia                                                               }
{===========================================================================}
Procedure InitInertia;

Var
  Func : Pointer;

Begin
  { TMT seems to have severe probles with floating point }
  { Let us initialize the FPU outselves.                 }
  Asm
    finit
  End;

  { Init the shader function array }
  Shaders[AMBIENT]  := ShadeAmbient;
  Shaders[LAMBERT]  := ShadeLambert;
  Shaders[GOURAUD]  := ShadeGouraud;
  Shaders[PHONG]    := ReflectionMap;
  Shaders[UNSHADED] := ShadeNull;

  { Init the polygon fillers function array }
  TriFillers[UNSHADED or SMOOTH] := UnshadedPolygon;
  TriFillers[AMBIENT  or SMOOTH] := AmbientPolygon;
  TriFillers[LAMBERT  or SMOOTH] := LambertPolygon;
  TriFillers[GOURAUD  or SMOOTH] := GouraudPolygon;
  TriFillers[PHONG    or SMOOTH] := PhongPolygon;

  TriFillers[UNSHADED or TEXTURE] := UnshadedTexturePolygon;
  TriFillers[AMBIENT  or TEXTURE] := AmbientTexturePolygon;
  TriFillers[LAMBERT  or TEXTURE] := LambertTexturePolygon;
  TriFillers[GOURAUD  or TEXTURE] := GouraudTexturePolygon;
  TriFillers[PHONG    or TEXTURE] := PhongTexturePolygon;

  TriFillers[UNSHADED or PERSPTEXTURE] := UnshadedPerspTexturePolygon;
  TriFillers[AMBIENT  or PERSPTEXTURE] := AmbientPerspTexturePolygon;
  TriFillers[LAMBERT  or PERSPTEXTURE] := LambertPerspTexturePolygon;
  TriFillers[GOURAUD  or PERSPTEXTURE] := GouraudPerspTexturePolygon;
  TriFillers[PHONG    or PERSPTEXTURE] := PhongPerspTexturePolygon;

  TriFillers[UNSHADED or REFLECTION] := UnshadedReflectionPolygon;
  TriFillers[AMBIENT  or REFLECTION] := AmbientReflectionPolygon;
  TriFillers[LAMBERT  or REFLECTION] := LambertReflectionPolygon;
  TriFillers[GOURAUD  or REFLECTION] := GouraudReflectionPolygon;
  TriFillers[PHONG    or REFLECTION] := PhongReflectionPolygon;

  TriFillers[TRANSPARENT or UNSHADED or SMOOTH] := TransUnshadedPolygon;
  TriFillers[TRANSPARENT or AMBIENT or SMOOTH] := TransAmbientPolygon;
  TriFillers[TRANSPARENT or LAMBERT or SMOOTH] := TransLambertPolygon;
  TriFillers[TRANSPARENT or GOURAUD or SMOOTH] := TransGouraudPolygon;
  TriFillers[TRANSPARENT or PHONG   or SMOOTH] := TransPhongPolygon;

  TriFillers[TRANSPARENT or UNSHADED or TEXTURE] := TransUnshadedTexturePolygon;
  TriFillers[TRANSPARENT or AMBIENT or TEXTURE] := TransAmbientTexturePolygon;
  TriFillers[TRANSPARENT or LAMBERT or TEXTURE] := TransLambertTexturePolygon;
  TriFillers[TRANSPARENT or GOURAUD or TEXTURE] := TransGouraudTexturePolygon;
  TriFillers[TRANSPARENT or PHONG   or TEXTURE] := TransPhongTexturePolygon;

  TriFillers[TRANSPARENT or UNSHADED or PERSPTEXTURE] := TransUnshadedPerspTexturePolygon;
  TriFillers[TRANSPARENT or AMBIENT  or PERSPTEXTURE] := TransAmbientPerspTexturePolygon;
  TriFillers[TRANSPARENT or LAMBERT  or PERSPTEXTURE] := TransLambertPerspTexturePolygon;
  TriFillers[TRANSPARENT or GOURAUD  or PERSPTEXTURE] := TransGouraudPerspTexturePolygon;
  TriFillers[TRANSPARENT or PHONG    or PERSPTEXTURE] := TransPhongPerspTexturePolygon;

  TriFillers[TRANSPARENT or UNSHADED or REFLECTION] := TransUnshadedReflectionPolygon;
  TriFillers[TRANSPARENT or AMBIENT or REFLECTION] := TransAmbientReflectionPolygon;
  TriFillers[TRANSPARENT or LAMBERT or REFLECTION] := TransLambertReflectionPolygon;
  TriFillers[TRANSPARENT or GOURAUD or REFLECTION] := TransGouraudReflectionPolygon;
  TriFillers[TRANSPARENT or PHONG   or REFLECTION] := TransPhongReflectionPolygon;

  { Initialize a few variables }
  AspectRatio := 1.2 / 1.3333333;

  FillChar(GlobalPalette, sizeof(GlobalPalette), 0);

  IntensityTable    := NIL;
  TransparencyTable := NIL;

  RenderList.SortList  := NIL;
  RenderList.Allocated := 0;
  RenderList.Count     := 0;

  Lights.NumLights     := 0;

  InitVertexPool;
  InitPolygonPool;
End;

{===========================================================================}
{ ShutDownInertia                                                           }
{===========================================================================}
Procedure ShutDownInertia;

Begin
  if (RenderList.SortList <> NIL)
    Then Free(RenderList.SortList);

  FreeVertexPool;
  FreePolygonPool;

  Asm
    finit
  End;
End;

{===========================================================================}
{ InitVectorObject                                                          }
{===========================================================================}
Procedure InitVectorObject(Var Obj : VectorObject);

Begin
  Obj.NumVert      := 0;
  Obj.NumPoly      := 0;
  Obj.Vertexs      := NIL;
  Obj.Polygons     := NIL;

  Obj.Cull         := 1;
  Obj.RenderMask   := AMBIENT or SMOOTH;
  Obj.Radius       := 0;
  Obj.CorrectCount := 0;

  InitMatrix(Obj.Orientation);
End;

{===========================================================================}
{ AllocObject                                                               }
{===========================================================================}
Function  AllocObject(Var Obj : VectorObject; NumVert, NumPoly : Longint) : Longint;

Begin
  InitVectorObject(Obj);

  Obj.NumVert := NumVert;
  Obj.NumPoly := NumPoly;

  Obj.Vertexs := Malloc(Obj.NumVert * sizeof(Vertex));
  if (Obj.Vertexs = NIL)
    Then Begin
      AllocObject := I_MEMERR;
      Exit;
    End;

  Obj.Polygons := Malloc(Obj.NumPoly * sizeof(Triangle));
  if (Obj.Polygons = NIL)
    Then Begin
      Free(Obj.Vertexs);
      AllocObject := I_MEMERR;
      Exit;
    End;

  AllocObject := I_OK;
End;

{===========================================================================}
{ FreeVectorObject                                                          }
{===========================================================================}
Procedure FreeVectorObject(Var Obj : VectorObject);

Begin
  if (Obj.Vertexs  <> NIL)
    Then Free(Obj.Vertexs);
  if (Obj.Polygons <> NIL)
    Then Free(Obj.Polygons);

  Obj.NumVert := 0;
  Obj.NumPoly := 0;
  Obj.Radius  := 0;
End;

{===========================================================================}
{ CopyVectorObject                                                          }
{===========================================================================}
Function  CopyVectorObject(Var Source, Dest : VectorObject) : Longint;

Begin
  if (AllocObject(Dest, Source.NumVert, Source.NumPoly) <> I_OK)
    Then Begin
      CopyVectorObject := I_MEMERR;
      Exit;
    End;

  Move(Source.Vertexs^,  Dest.Vertexs^,  Source.NumVert * sizeof(Vertex));
  Move(Source.Polygons^, Dest.Polygons^, Source.NumPoly * sizeof(Triangle));

  MatrixCopy(Source.Orientation, Dest.Orientation);

  Dest.RenderMask   := Source.RenderMask;
  Dest.Radius       := Source.Radius;
  Dest.CorrectCount := Source.CorrectCount;
  Dest.ClipMask     := Source.ClipMask;
  Dest.Cull         := Source.Cull;

  CopyVectorObject := I_OK;
End;

{===========================================================================}
{ CalcBoundartVolume                                                        }
{===========================================================================}
Procedure CalcBoundaryVolume(Var Obj : VectorObject);

Var
  Error : Single;
  Temp  : Single;
  x, y, z : Single;
  i     : Longint;

Begin
  if (Obj.NumVert = 0) Then Exit;

  Error := 0;
  For i := 0 to Obj.NumVert - 1 do
    Begin
      x := Obj.Vertexs^[i].Loc.x;
      y := Obj.Vertexs^[i].Loc.y;
      z := Obj.Vertexs^[i].Loc.z;

      Temp := SQR(x) + SQR(y) + SQR(z);

      if (Temp > Error) Then Error := Temp;
    End;

  Obj.Radius := sqrt(Error);
End;

{===========================================================================}
{ LoadGVO                                                                   }
{===========================================================================}
Function  LoadGVO(Var Obj : VectorObject; Filename : String; FOfs : Longint) : Longint;

Var
  fp       : pFile;
  GVOTag   : Array[0..2] of Char;
  Flag     : Integer;
  SInt     : Integer;
  DiskPos  : Longint;
  i        : Longint;
  TempChar : Byte;

Begin
  InitVectorObject(Obj);

  fp := FOpen(Filename, 'r');

  if (fp = NIL)
    Then Begin
      LoadGVO := I_FILEERR;
      Exit;
    End;

  fseek(fp, FOfs, SEEK_SET);

  fread(GVOTag, sizeof(GVOTag), 1, fp);

  if ((GVOTag[0] <> 'G') And (GVOTag[1] <> 'V')) And (GVOTag[2] <> 'O')
    Then Begin
      FClose(fp);
      LoadGVO := I_FORMATERR;
      Exit;
    End;

  fread(Flag, sizeof(Flag), 1, fp);

  if (((Flag And $01) <> 0) And ((Flag And $02) <> 0))
    Then Begin
      fread(SInt, sizeof(SInt), 1, fp);
      Obj.NumVert := SInt;
      DiskPos := ftell(fp);
      fseek(fp, DiskPos + (Obj.NumVert * 6), SEEK_SET);
      fread(SInt, sizeof(SInt), 1, fp);
      Obj.NumPoly := SInt;
      fseek(fp, DiskPos, SEEK_SET);
    End
    Else Begin
      fclose(fp);
      LoadGVO := I_FILEERR;
      Exit;
    End;

  if (AllocObject(Obj, Obj.NumVert, Obj.NumPoly) <> I_OK)
    Then Begin
      fclose(fp);
      LoadGVO := I_MEMERR;
      Exit;
    End;

  if ((Flag And 1) <> 0)
    Then Begin
      For i := 0 to Obj.NumVert - 1 do
        Begin
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Vertexs^[i].Loc.x := SInt;

          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Vertexs^[i].Loc.y := SInt;
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Vertexs^[i].Loc.z := SInt;
        End;
    End;

  if ((Flag And 2) <> 0)
    Then Begin
      fread(SInt, sizeof(SInt), 1, fp);

      For i := 0 to Obj.NumPoly - 1 do
        Begin
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Polygons^[i].A := @(Obj.Vertexs^[SInt]);
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Polygons^[i].B := @(Obj.Vertexs^[SInt]);
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Polygons^[i].C := @(Obj.Vertexs^[SInt]);
        End;
    End;

  if ((Flag And 4) <> 0)
    Then Begin
      For i := 0 to Obj.NumPoly - 1 do
        Begin
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Polygons^[i].N.x := SInt / 256.0;
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Polygons^[i].N.y := SInt / 256.0;
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Polygons^[i].N.z := SInt / 256.0;
        End;
    End;

  if ((Flag And 8) <> 0)
    Then Begin
      For i := 0 to Obj.NumVert - 1 do
        Begin
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Vertexs^[i].N.x := SInt / 256.0;
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Vertexs^[i].N.y := SInt / 256.0;
          fread(SInt, sizeof(SInt), 1, fp);
          Obj.Vertexs^[i].N.z := SInt / 256.0;
        End;
    End;

  if ((Flag And 16) <> 0)
    Then Begin
      For i := 0 to Obj.NumPoly - 1 do
        Begin
          fread(TempChar, sizeof(TempChar), 1, fp);
          Obj.Polygons^[i].UV1.x := TempChar / 255.0;
          fread(TempChar, sizeof(TempChar), 1, fp);
          Obj.Polygons^[i].UV1.y := TempChar / 255.0;

          fread(TempChar, sizeof(TempChar), 1, fp);
          Obj.Polygons^[i].UV2.x := TempChar / 255.0;
          fread(TempChar, sizeof(TempChar), 1, fp);
          Obj.Polygons^[i].UV2.y := TempChar / 255.0;

          fread(TempChar, sizeof(TempChar), 1, fp);
          Obj.Polygons^[i].UV3.x := TempChar / 255.0;
          fread(TempChar, sizeof(TempChar), 1, fp);
          Obj.Polygons^[i].UV3.y := TempChar / 255.0;
        End;
    End;

  fclose(fp);

  CalcBoundaryVolume(Obj);

  LoadGVO := I_OK;
End;

{===========================================================================}
{ SetAbsoluteRotation                                                       }
{===========================================================================}
Procedure SetAbsoluteRotation(Var Obj : VectorObject; rx, ry, rz : Single);

Begin
  SetRotationMatrix(Obj.Orientation, rx, ry, rz);
  Obj.CorrectCount := 0;
End;

{===========================================================================}
{ SetDeltaRotation                                                          }
{===========================================================================}
Procedure SetDeltaRotation(Var Obj : VectorObject; dx, dy, dz : Single);

Var
  TempMat  : Matrix4x3;
  TempMat2 : Matrix4x3;

Begin
  SetRotationMatrix(TempMat, dx, dy, dz);
  MatrixCopy(Obj.Orientation, TempMat2);
  MatrixMul(Obj.Orientation, TempMat, TempMat2);

  Inc(Obj.CorrectCount);

  if (Obj.CorrectCount >= CORRECTRES)
    Then Begin
      MatrixOrthoNormalize(Obj.Orientation);
      Obj.CorrectCount := 0;
    End;
End;

{===========================================================================}
{ SetAbsoluteLocation                                                       }
{===========================================================================}
Procedure SetAbsoluteLocation(Var Obj : VectorObject; tx, ty, tz : Single);

Begin
  Obj.Orientation[XLOC] := tx;
  Obj.Orientation[YLOC] := ty;
  Obj.Orientation[ZLOC] := tz;
End;

{===========================================================================}
{ SetDeltaLocation                                                          }
{===========================================================================}
Procedure SetDeltaLocation(Var Obj : VectorObject; dx, dy, dz : Single);

Begin
  Obj.Orientation[XLOC] := Obj.Orientation[XLOC] + dx;
  Obj.Orientation[YLOC] := Obj.Orientation[YLOC] + dy;
  Obj.Orientation[ZLOC] := Obj.Orientation[ZLOC] + dz;
End;

{===========================================================================}
{ SetBaseColor                                                              }
{===========================================================================}
Procedure SetBaseColor(Var Obj : VectorObject; C : Byte);

Var
  i : Longint;

Begin
  For i := 0 to Obj.NumPoly - 1 do
    Obj.Polygons^[i].Color := C;
End;

{===========================================================================}
{ SetPolygonColor                                                           }
{===========================================================================}
Procedure SetPolygonColor(Var Obj : VectorObject; Poly : Longint; C : Byte);

Begin
  Obj.Polygons^[Poly].Color := C;
End;

{===========================================================================}
{ SetObjectTexture                                                          }
{===========================================================================}
Procedure SetObjectTexture(Var Obj : VectorObject; C : Longint);

Var
  i : Longint;

Begin
  for i := 0 to Obj.NumPoly - 1 do
    Obj.Polygons^[i].Texture := TextureList[C].Texture;
End;

{===========================================================================}
{ SetPolygonTexture                                                         }
{===========================================================================}
Procedure SetPolygonTexture(Var Obj : VectorObject; Poly, C : Longint);

Begin
  Obj.Polygons^[Poly].Texture := TextureList[C].Texture;
End;

{===========================================================================}
{ SetObjectPhongMap                                                         }
{===========================================================================}
Procedure SetObjectPhongMap(Var Obj : VectorObject; C : Longint);

Var
  i : Longint;

Begin
  for i := 0 to Obj.NumPoly - 1 do
    Obj.Polygons^[i].IBuffer:= TextureList[C].Texture;
End;

{===========================================================================}
{ SetPolygonPhongMap                                                        }
{===========================================================================}
Procedure SetPolygonPhongMap(Var Obj : VectorObject; Poly, C : Longint);

Begin
  Obj.Polygons^[Poly].IBuffer := TextureList[C].Texture;
End;

{===========================================================================}
{ SetRenderMask                                                             }
{===========================================================================}
Procedure SetRenderMask(Var Obj : VectorObject; Mask : Longint);

Begin
  Obj.RenderMask := Mask;
  Obj.ClipMask   := 0;

  Case (Obj.RenderMask And SURFACE_MASK) of
    TEXTURE      : Obj.ClipMask := Obj.ClipMask or CLIP_TEXTURE;
    PERSPTEXTURE : Obj.ClipMask := Obj.ClipMask or CLIP_TEXTURE;
    REFLECTION   : Obj.ClipMask := Obj.ClipMask or CLIP_ENV;
  End;

  Case (Obj.RenderMask And SHADE_MASK) of
    GOURAUD : Obj.ClipMask := Obj.ClipMask or CLIP_GOURAUD;
    PHONG   : Obj.ClipMask := Obj.ClipMask or CLIP_ENV;
  End;
End;

{===========================================================================}
{ SetCull                                                                   }
{===========================================================================}
Procedure SetCull(Var Obj : VectorObject; Cull : Longint);

Begin
  Obj.Cull := Cull;
End;

{===========================================================================}
{ SetAspectRatio                                                            }
{===========================================================================}
Procedure SetAspectRatio(ARatio : Single);

Begin
  AspectRatio := ARatio;
End;

{===========================================================================}
{ SetFocalLength                                                            }
{===========================================================================}
Procedure SetFocalLength(Var View : ViewObject; Fl : Single);

Begin
  View.FocalLength := Fl;
  SetFieldOfView(View, (15.0 / Fl) * 160.0);
End;

{===========================================================================}
{ SetFieldOfView                                                            }
{===========================================================================}
Procedure SetFieldOfView(Var View : ViewObject; FOV : Single);

Begin
  View.FieldOfView := FOV;
  View.CamZFac     := tan(DEG2RAD(View.FieldOfView / 2.0));
  View.InvCamZFac  := 1.0 / View.CamZFac;
End;

{===========================================================================}
{ InitView                                                                  }
{===========================================================================}
Procedure InitView(Var View : ViewObject);

Var
  TimerAddr : DWord;

Begin
  SetViewPort(View, 0, 0, 319, 199, True);
  InitMatrix(View.Camera);
  SetFieldOfView(View, 65);
  SetViewFustrum(View, 1.0, 100000.0);
End;

{===========================================================================}
{ FreeView                                                                  }
{===========================================================================}
Procedure FreeView(Var View : ViewObject);

Begin
End;


{===========================================================================}
{ GetObjectAsCamera                                                         }
{===========================================================================}
Procedure GetObjectAsCamera(Var View : ViewObject; Obj : VectorObject);

Begin
  MatrixCopy(View.Camera, Obj.Orientation);
End;

{===========================================================================}
{ SetViewFustrum                                                            }
{===========================================================================}
Procedure SetViewFustrum(Var View : ViewObject; MinZ, MaxZ : Single);

Begin
  View.FustrumMinZ := MinZ;
  View.FustrumMaxZ := MaxZ;
End;

{===========================================================================}
{ SetViewPort                                                               }
{===========================================================================}
Procedure SetViewPort(Var View : ViewObject; X1, Y1, X2, Y2 : Single; Update : Boolean);

Begin
  if (X1 > X2)
    Then SingleSWAP(X1, X2);
  if (Y1 > Y2)
    Then SingleSWAP(Y1, Y2);

  View.WinMinX := Round(X1);
  View.WinMinY := Round(Y1);
  View.WinMaxX := Round(X2);
  View.WinMaxY := Round(Y2);

  View.CamYFac    := ((X2 - X1 + 1) / (Y2 - Y1 + 1)) * AspectRatio;
  View.InvCamYFac := 1.0 / View.CamYFac;

  if (Update)
    Then Begin
      View.CenterX := (X2 + X1) / 2.0;
      View.CenterY := (Y2 + Y1) / 2.0;
    End;
End;

{===========================================================================}
{ SetViewLocation                                                           }
{===========================================================================}
Procedure SetViewLocation(Var View : ViewObject; X, Y, Z : Single);

Begin
  View.Camera[XLOC] := X;
  View.Camera[YLOC] := Y;
  View.Camera[ZLOC] := Z;
End;

{===========================================================================}
{ SetFocusPoint                                                             }
{===========================================================================}
Procedure SetFocusPoint(Var View : ViewObject; TargetX, TargetY, TargetZ, Roll : Single);

Var
  ViewMat : Matrix4x3;
  RollMat : Matrix4x3;
  zx, zy, zz : Single;
  yx, yy, yz : Single;
  xx, xy, xz : Single;
  gx, gy, gz : Single;
  Len        : Single;

Begin
  zx := TargetX - View.Camera[XLOC];
  zy := TargetY - View.Camera[YLOC];
  zz := TargetZ - View.Camera[ZLOC];

  if ((zx = 0) And (zy = 0)) And (zz = 0)
    Then Exit;

  Len := sqrt(SQR(zx) + SQR(zy) + SQR(zz));

  zx := (zx/Len);
  zy := (zy/Len);
  zz := (zz/Len);

  if ((zx = 0) And (zz = 0))
    Then Begin
      gx := -zy;
      gy := 0;
      gz := 0;
    End
    Else Begin
      gx := 0;
      gy := -1;
      gz := 0;
    End;

  (* cross G vector with Z vector *)
  xx := gy * zz - gz * zy;
  xy := gz * zx - gx * zz;
  xz := gx * zy - gy * zx;
  Len := sqrt(SQR(xx) + SQR(xy) + SQR(xz));
  xx := xx/Len;
  xy := xy/Len;
  xz := xz/Len;

  yx := xy*zz - xz*zy;
  yy := xz*zx - xx*zz;
  yz := xx*zy - xy*zx;
  Len := sqrt(SQR(yx) + SQR(yy) + SQR(yz));
  yx := yx/Len;
  yy := yy/Len;
  yz := yz/Len;

  ViewMat[0] := xx;  ViewMat[1] := yx;  ViewMat[2] := zx;
  ViewMat[3] := xy;  ViewMat[4] := yy;  ViewMat[5] := zy;
  ViewMat[6] := xz;  ViewMat[7] := yz;  ViewMat[8] := zz;

  SetRotationMatrix(RollMat, 0, 0, Roll);

  MatrixMul(View.Camera, ViewMat, RollMat);
End;

{===========================================================================}
{ PushLightSource                                                           }
{===========================================================================}
Procedure PushLightSource(X, Y, Z : Single);

Var
  Index : Longint;
  Len   : Single;

Begin
  if (Lights.NumLights >= MAXLIGHTS)
    Then Exit;

  Index := Lights.NumLights;

  Len := 1.0 / sqrt(sqr(x) + sqr(y) + sqr(z));

  Lights.Light[Index].x := X * Len;
  Lights.Light[Index].y := Y * Len;
  Lights.Light[Index].z := Z * Len;

  Inc(Lights.NumLights);
End;

{===========================================================================}
{ PopLightSource                                                            }
{===========================================================================}
Procedure PopLightSource;

Begin
  if (Lights.NumLights > 0)
    Then Dec(Lights.NumLights);
End;

{===========================================================================}
{ RotateLightSource                                                         }
{===========================================================================}
Procedure RotateLightSource(Index : Longint; Xa, Ya, Za : Single);

Begin
  SingleRotate(Lights.Light[Index], Xa, Ya, Za);
End;

{===========================================================================}
{ ByteSort                                                                  }
{===========================================================================}
Procedure ByteSort;

Var
  i, Index : Longint;
  TempPtr  : pSortRec;

Begin
  For i := 0 to 255 do
    Begin
      RenderList.LSBList[i] := NIL;
      RenderList.MSBList[i] := NIL;
    End;

  For i := 0 to RenderList.Count - 1 do
    Begin
      Index := RenderList.SortList^[i].ZMid And $ff;
      RenderList.SortList^[i].Next := RenderList.LSBList[Index];
      RenderList.LSBList[Index] := @RenderList.SortList^[i];
    End;

  For i := 0 to 255 do
    Begin
      while (RenderList.LSBList[i] <> NIL) do
        Begin
          Index   := (RenderList.LSBList[i]^.ZMid Shr 8) And $ff;
          TempPtr := RenderList.MSBList[Index];
          RenderList.MSBList[Index] := RenderList.LSBList[i];
          RenderList.LSBList[i]     := RenderList.LSBList[i]^.Next;
          RenderList.MSBList[Index]^.Next := TempPtr;
        End;
    End;
End;

{===========================================================================}
{ Render                                                                    }
{===========================================================================}
Procedure Render(VPage : Pointer; X, Y : Longint);

Var
  i : Longint;

Begin
  ByteSort;

  For i := 255 downto 0 do
    Begin
      While (RenderList.MSBList[i] <> NIL) do
        Begin
          TriFillers[RenderList.MSBList[i]^.ShadeSurface](RenderList.MSBList[i]^.Poly, VPage, X, Y);
          RenderList.MSBList[i] := RenderList.MSBList[i]^.Next;
        End;
    End;

  RenderList.Count := 0;
  ResetVertexPool;
  ResetPolygonPool;
End;

{===========================================================================}
{ ObjectVisible                                                             }
{===========================================================================}
Function ObjectVisible(View : ViewObject; Var Obj : VectorObject; M : Matrix4x3; Dot : Single) : Boolean;

Var
  box_x1, box_x2 : Single;
  box_y1, box_y2 : Single;
  box_z1, box_z2 : Single;

Begin
  Obj.ClipMask := Obj.ClipMask And CLIP_SHADEMASK;

  { build a basic bounding box }
  box_z1 := (M[ZLOC] - Obj.Radius) * View.CamZFac;
  box_z2 := (M[ZLOC] + Obj.Radius) * View.CamZFac;
  box_x1 := M[XLOC] - Obj.Radius;
  box_x2 := M[XLOC] + Obj.Radius;
  box_y1 := (M[YLOC] - Obj.Radius) * View.CamYFac;
  box_y2 := (M[YLOC] + Obj.Radius) * View.CamYFac;

  { complete rejection tests }
  If ((box_x2 < -box_z1) And (box_x2 < -box_z2))
    Then Begin
      ObjectVisible := False;
      Exit;
    End;

  If ((box_x1 > box_z1)  And (box_x1 > box_z2))
    Then Begin
      ObjectVisible := False;
      Exit;
    End;

  If ((box_y2 < -box_z1) And (box_y2 < -box_z2))
    Then Begin
      ObjectVisible := False;
      Exit;
    End;

  If ((box_y1 > box_z1)  And (box_y1 > box_z2))
    Then Begin
      ObjectVisible := False;
      Exit;
    End;

  If ((M[ZLOC] + (Obj.Radius * View.CamZFac)) < View.FustrumMinZ)
    Then Begin
      ObjectVisible := False;
      Exit;
    End;

  { clipping tests }

  If (box_z1 < View.FustrumMinZ)
    Then Obj.ClipMask := Obj.ClipMask or CLIP_Z;

  If (box_x1 < -box_z1)
    Then Obj.ClipMask := Obj.ClipMask or CLIP_LEFT;

  If (box_x2 > box_z1)
    Then Obj.ClipMask := Obj.ClipMask or CLIP_RIGHT;

  If (box_y1 < -box_z1)
    Then Obj.ClipMask := Obj.ClipMask or CLIP_BOTTOM;

  If (box_y2 > box_z1)
    Then Obj.ClipMask := Obj.ClipMask or CLIP_TOP;

  ObjectVisible := True;
End;

{===========================================================================}
{ AddToRenderList                                                           }
{===========================================================================}
Procedure AddToRenderList(View : ViewObject; Var Obj : VectorObject);

Var
  CullVector : Vertex3d;
  TempVec    : Vertex3d;
  FustrumZ   : Vertex3d;
  TempLoc    : Vertex3d;
  TempMat    : Matrix4x3;
  ScaleMat   : Matrix4x3;
  TempCamera : Matrix4x3;
  Dot        : Single;
  i          : Longint;

Begin
  Asm
    ffree st(0)
    ffree st(1)
    ffree st(2)
    ffree st(3)
    ffree st(4)
    ffree st(5)
    ffree st(6)
    ffree st(7)
  End;

  TempVec.x := Obj.Orientation[XLOC] - View.Camera[XLOC];
  TempVec.y := Obj.Orientation[YLOC] - View.Camera[YLOC];
  TempVec.z := Obj.Orientation[ZLOC] - View.Camera[ZLOC];

  FustrumZ.x := View.Camera[2];
  FustrumZ.y := View.Camera[5];
  FustrumZ.z := View.Camera[8];

  Dot := DotProduct(TempVec, FustrumZ);

  TempMat[XLOC] := View.Camera[0] * TempVec.x + View.Camera[3] * TempVec.y + View.Camera[6] * TempVec.z;
  TempMat[YLOC] := View.Camera[1] * TempVec.x + View.Camera[4] * TempVec.y + View.Camera[7] * TempVec.z;
  TempMat[ZLOC] := View.Camera[2] * TempVec.x + View.Camera[5] * TempVec.y + View.Camera[8] * TempVec.z;

  if (Not(ObjectVisible(View, Obj, TempMat, Dot)))
    Then Exit;

  MatrixMul(TempMat, View.Camera, Obj.Orientation);

  CullVector.x := -TempMat[XLOC] * View.InvCamZFac;
  CullVector.y := -TempMat[YLOC] * View.InvCamZFac;
  CullVector.z := -TempMat[ZLOC] * View.InvCamZFac;

  if ((CullVector.x = 0) And (CullVector.y = 0)) And (CullVector.z = 0)
    Then Begin
      CullVector.x := -TempMat[2];
      CullVector.y := -TempMat[5];
      CullVector.z := -TempMat[8];
    End
    Else Begin
      InvXForm(CullVector, TempMat);
    End;

  for i := 0 to Lights.NumLights - 1 do
    Begin
      Lights.TempLight[i] := Lights.Light[i];
      InvXForm(Lights.TempLight[i], Obj.Orientation);
    End;


  if (Obj.Cull = 1)
    Then Begin
      for i := 0 to Obj.NumVert - 1 do
        Obj.Vertexs^[i].Visible := 1;

      CullPolys(Obj, CullVector);
    End
    Else Begin
      for i := 0 to Obj.NumVert - 1 do
        Obj.Vertexs^[i].Visible := 1;

      for i := 0 to Obj.NumPoly - 1 do
        Obj.Polygons^[i].Visible := 1;
    End;

  SetScaleMatrix(ScaleMat, 1.0, View.CamYFac, View.CamZFac);
  MatrixCopy(TempMat, TempCamera);
  MatrixMul(TempCamera, TempMat, ScaleMat);

  ShadeXForm(View, Obj, TempCamera);
End;

{===========================================================================}
{ LoadIntensityTable                                                        }
{===========================================================================}
Function LoadIntensityTable(Filename : String; FOfs : Longint) : Longint;

Var
  fp : pFile;
  i : Longint;

Begin
  IntensityTable := Malloc(256 * 64);

  if (IntensityTable = NIL)
    Then Begin
      LoadIntensityTable := I_MEMERR;
      Exit;
    End;

  fp := FOpen(Filename, 'r');
  if (fp = NIL)
    Then Begin
      Free(IntensityTable);
      LoadIntensityTable := I_FILEERR;
      Exit;
    End;

  FSeek(fp, FOfs, SEEK_SET);
  For i := 63 downto 0 do
    Begin
      FRead(IntensityTable^[i*256], 1, 256, fp);
    End;

  FClose(fp);
  LoadIntensityTable := I_OK;
End;

{===========================================================================}
{ FreeIntensityTable                                                        }
{===========================================================================}
Procedure FreeIntensityTable;

Begin
  If (IntensityTable <> Nil)
    Then Free(IntensityTable);
End;

{===========================================================================}
{ LoadTransparencyTable                                                        }
{===========================================================================}
Function LoadTransparencyTable(Filename : String; FOfs : Longint) : Longint;

Var
  fp : pFile;
  i : Longint;

Begin
  TransparencyTable := Malloc(256 * 256);

  if (TransparencyTable = NIL)
    Then Begin
      LoadTransparencyTable := I_MEMERR;
      Exit;
    End;

  fp := FOpen(Filename, 'r');
  if (fp = NIL)
    Then Begin
      Free(TransparencyTable);
      LoadTransparencyTable := I_FILEERR;
      Exit;
    End;

  FSeek(fp, FOfs, SEEK_SET);

  FRead(TransparencyTable^, 1, 256 * 256, fp);

  FClose(fp);
  LoadTransparencyTable := I_OK;
End;

{===========================================================================}
{ FreeTransparncyTable                                                      }
{===========================================================================}
Procedure FreeTransparencyTable;

Begin
  If (TransparencyTable <> Nil)
    Then Free(TransparencyTable);
End;

{===========================================================================}
{ LoadPalette                                                               }
{===========================================================================}
Function LoadPalette(FileName : String; FOfs : Longint) : Longint;

Var
  fp : pFile;

Begin
  fp := FOpen(Filename, 'r');
  If (FP = NIL)
    Then Begin
      LoadPalette := I_FILEERR;
      Exit;
    End;

  FSeek(fp, FOfs, SEEK_SET);
  FRead(GlobalPalette, 1, Sizeof(Palette), fp);

  FClose(fp);
  LoadPalette := I_OK;
End;

{===========================================================================}
{ LoadPCXTexture                                                            }
{===========================================================================}
Function  LoadPCXTexture(Index : Longint; Filename : String; FOfs : Longint) : Longint;

Var
  T : Longint;
  RetCode : Longint;

Begin
  RetCode := LoadPCXFile(Filename, TextureList[Index].Texture, TextureList[Index].Pal, T, FOfs);

  Case RetCode of
    0 :  Begin LoadPCXTexture := I_OK; Exit; End;
    Else Begin LoadPCXTexture := I_FILEERR;  Exit; End;
  End;
End;

{===========================================================================}
{ FreeTexture                                                               }
{===========================================================================}
Procedure FreeTexture(Index : Longint);

Begin
  if (TextureList[Index].Texture <> NIL)
    Then Free(TextureList[Index].Texture);
End;

{===========================================================================}
{ GetTexturePalette                                                         }
{===========================================================================}
Procedure GetTexturePalette(Index : Longint);

Begin
  Move(TextureList[Index].Pal, GlobalPalette, sizeof(GlobalPalette));
End;

{===========================================================================}
{ MakePhongMap                                                              }
{===========================================================================}
Function MakePhongMap(Index : Longint) : Longint;

Var
  t, x, y : Longint;
  tx, ty : Single;
  Adr : DWord;

Begin
  TextureList[Index].Texture := Malloc(256 * 256);
  if (TextureList[Index].Texture = Nil)
    Then Begin
      MakePhongMap := I_MEMERR;
      Exit;
    End;

  Adr := DWord(TextureList[Index].Texture);

  For y := 0 to 255 do
    For x := 0 to 255 do
      Begin
        tx := 127.0 - x;
        ty := 127.0 - y;
        t := Round(Sqrt(sqr(tx) + Sqr(ty)) / 2);
        if (t < 0)  Then t := 0;
        if (t > 63) Then t := 63;
        Mem[Adr] := Byte(63 - t);
        Inc(Adr);
      End;

  MakePhongMap := I_OK;
End;

Begin
  InitInertia;
End.
